//------------------------------------------------------------------------------
//
// Copyright (c) Microsoft Corporation.
// All rights reserved.
//
// This code is licensed under the MIT License.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files(the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions :
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
//------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using CoreFoundation;
using Foundation;
using Microsoft.IdentityModel.Clients.ActiveDirectory.Internal.Flows;
using UIKit;
using Microsoft.Identity.Core;
using Microsoft.IdentityModel.Clients.ActiveDirectory.Internal.Helpers;
using Microsoft.IdentityModel.Clients.ActiveDirectory.Internal.OAuth2;
using Microsoft.Identity.Core.Cache;
using Microsoft.Identity.Core.Helpers;
using Microsoft.IdentityModel.Clients.ActiveDirectory.Internal.Broker;
using System.Globalization;
using Security;

namespace Microsoft.IdentityModel.Clients.ActiveDirectory.Internal.Platform
{
    internal class iOSBroker : IBroker
    {
        private static SemaphoreSlim brokerResponseReady = null;
        private static NSUrl brokerResponse = null;

        private readonly ICoreLogger _logger;

        private string _brokerRequestNonce;
        private bool _brokerV3Installed = false;

        public iOSBroker(ICoreLogger logger)
        {
            _logger = logger;
        }

        public IPlatformParameters PlatformParameters { get; set; }

        public bool CanInvokeBroker
        {
            get
            {
                PlatformParameters pp = PlatformParameters as PlatformParameters;
                if (pp == null)
                {
                    return false;
                }

                bool canStartBroker = false;

                if (pp.UseBroker)
                {
                    pp.CallerViewController.InvokeOnMainThread(() =>
                    {
                        if (IsBrokerInstalled("msauthv3://"))
                        {
                            _logger.Info("iOS Broker (msauthv3://) can be invoked. ");
                            _brokerV3Installed = true;
                            canStartBroker = true;
                        }
                    });

                    if (!canStartBroker)
                    {
                        pp.CallerViewController.InvokeOnMainThread(() =>
                        {
                            if (IsBrokerInstalled("msauth://"))
                            {
                                _logger.Info("iOS Broker (msauth://) can be invoked. ");
                                canStartBroker = true;
                            }
                        });
                    }
                }

                return canStartBroker;
            }
        }

        private bool IsBrokerInstalled(string brokerUriScheme)
        {
            return UIApplication.SharedApplication.CanOpenUrl(new NSUrl(brokerUriScheme));
        }

        public async Task<AdalResultWrapper> AcquireTokenUsingBrokerAsync(IDictionary<string, string> brokerPayload)
        {
            if (brokerPayload.ContainsKey(BrokerParameter.SilentBrokerFlow))
            {
                _logger.Info("iOS Broker payload contains silent flow key in payload. " +
                    "Throwing AdalSilentTokenAcquisitionException() ");
                throw new AdalSilentTokenAcquisitionException();
            }

            brokerResponse = null;
            brokerResponseReady = new SemaphoreSlim(0);

            //call broker
            string base64EncodedString = Base64UrlHelpers.Encode(BrokerKeyHelper.GetOrCreateBrokerKey(_logger));
            brokerPayload["broker_key"] = base64EncodedString;
            brokerPayload["max_protocol_ver"] = "2";

            if (brokerPayload.TryGetValue("claims", out string claims) && !string.IsNullOrEmpty(claims))
            {
                brokerPayload.Add("skip_cache", "YES");
                string encodedClaims = EncodingHelper.UrlEncode(claims);
                brokerPayload[BrokerParameter.Claims] = encodedClaims;
            }

            if (_brokerV3Installed)
            {
                _brokerRequestNonce = Guid.NewGuid().ToString();
                brokerPayload[BrokerParameter.BrokerNonce] = _brokerRequestNonce;

                string applicationToken = TryReadBrokerApplicationTokenFromKeychain(brokerPayload);

                if (!string.IsNullOrEmpty(applicationToken))
                {
                    brokerPayload.Add(BrokerConstants.ApplicationToken, applicationToken);
                }
            }

            if (brokerPayload.ContainsKey(BrokerParameter.BrokerInstallUrl))
            {
                HandleInstallUrl(brokerPayload);
            }
            else
            {
                string queryParams = brokerPayload.ToQueryParameter();
                _logger.InfoPii("Invoking the iOS broker " + queryParams,
                    "Invoking the iOS broker. ");
                NSUrl url = new NSUrl("msauth://broker?" + queryParams);
                DispatchQueue.MainQueue.DispatchAsync(() => UIApplication.SharedApplication.OpenUrl(url));
            }

            await brokerResponseReady.WaitAsync().ConfigureAwait(false);
            PlatformParameters = null;
            return ProcessBrokerResponse();
        }

        private void HandleInstallUrl(IDictionary<string, string> brokerPayload)
        {
            string url = brokerPayload[BrokerParameter.BrokerInstallUrl];
            Uri uri = new Uri(url);
            string query = uri.Query;
            if (query.StartsWith("?", StringComparison.OrdinalIgnoreCase))
            {
                query = query.Substring(1);
            }

            _logger.Info("Invoking the iOS broker app link");

            Dictionary<string, string> keyPair = EncodingHelper.ParseKeyValueList(query, '&', true, false, null);
            DispatchQueue.MainQueue.DispatchAsync(() => UIApplication.SharedApplication.OpenUrl(new NSUrl(keyPair["app_link"])));

            throw new AdalException(AdalErrorIOSEx.BrokerApplicationRequired, AdalErrorMessageIOSEx.BrokerApplicationRequired);
        }

        private AdalResultWrapper ProcessBrokerResponse()
        {
            _logger.InfoPii(
                "Processing reponse from iOS broker " + brokerResponse.Query, 
                "Processing response from iOS Broker. ");

            string[] keyValuePairs = brokerResponse.Query.Split('&');

            IDictionary<string, string> responseDictionary = new Dictionary<string, string>();
            foreach (string pair in keyValuePairs)
            {
                string[] keyValue = pair.Split('=');
                responseDictionary[keyValue[0]] = EncodingHelper.UrlDecode(keyValue[1]);
                if (responseDictionary[keyValue[0]].Equals("(null)", StringComparison.OrdinalIgnoreCase) && keyValue[0].Equals("code", StringComparison.OrdinalIgnoreCase))
                {
                    responseDictionary["error"] = "broker_error";
                }
            }

            return ResultFromBrokerResponse(responseDictionary);
        }

        private AdalResultWrapper ResultFromBrokerResponse(IDictionary<string, string> responseDictionary)
        {
            TokenResponse response;

            if (responseDictionary.ContainsKey("error") || responseDictionary.ContainsKey("error_description"))
            {
                _logger.Info("Broker response returned an error. ");
                response = TokenResponse.CreateFromBrokerResponse(responseDictionary);
            }
            else
            {
                string expectedHash = responseDictionary["hash"];
                string encryptedResponse = responseDictionary["response"];
                string decryptedResponse = BrokerKeyHelper.DecryptBrokerResponse(encryptedResponse);
                string responseActualHash = PlatformProxyFactory.GetPlatformProxy().CryptographyManager.CreateSha256Hash(decryptedResponse);
                byte[] rawHash = Convert.FromBase64String(responseActualHash);
                string hash = BitConverter.ToString(rawHash);

                if (expectedHash.Equals(hash.Replace("-", ""), StringComparison.OrdinalIgnoreCase))
                {
                    responseDictionary = EncodingHelper.ParseKeyValueList(decryptedResponse, '&', false, null);
                    response = TokenResponse.CreateFromBrokerResponse(responseDictionary);
                    _logger.Info("Broker response successful. ");
                }
                else
                {
                    response = new TokenResponse
                    {
                        Error = AdalError.BrokerReponseHashMismatch,
                        ErrorDescription = AdalErrorMessage.BrokerReponseHashMismatch
                    };
                    _logger.InfoPii("Broker response hash mismatch: " + response.Error, "Broker response hash mismatch. ");
                }

                if (!ValidateBrokerResponseNonceWithRequestNonce(responseDictionary))
                {
                    response = new TokenResponse
                    {
                        Error = AdalError.BrokerNonceMismatch,
                        ErrorDescription = AdalErrorMessage.BrokerNonceMismatch
                    };
                }

                if (responseDictionary.ContainsKey(BrokerConstants.ApplicationToken))
                {
                    TryWriteBrokerApplicationTokenToKeychain(
                        responseDictionary[BrokerParameter.ClientId],
                        responseDictionary[BrokerConstants.ApplicationToken]);
                }
            }

            var dateTimeOffset = new DateTimeOffset(new DateTime(1970, 1, 1, 0, 0, 0, 0, DateTimeKind.Utc));
            dateTimeOffset = dateTimeOffset.AddSeconds(response.ExpiresOn);
            return response.GetResult(dateTimeOffset, dateTimeOffset);
        }

        private bool ValidateBrokerResponseNonceWithRequestNonce(IDictionary<string, string> brokerResponseDictionary)
        {
            if (_brokerV3Installed)
            {
                string brokerResponseNonce = brokerResponseDictionary[BrokerParameter.BrokerNonce];
                bool ok = string.Equals(brokerResponseNonce, _brokerRequestNonce, StringComparison.InvariantCultureIgnoreCase);

                if (!ok)
                {
                    _logger.Error(
                        string.Format(
                            CultureInfo.CurrentCulture,
                            "Nonce check failed! Broker response nonce is:  {0}, \nBroker request nonce is: {1}",
                            brokerResponseNonce,
                            _brokerRequestNonce));
                }
                return ok;
            }
            return true;
        }

        private void TryWriteBrokerApplicationTokenToKeychain(string clientId, string applicationToken)
        {
            iOSTokenCacheAccessor iOSTokenCacheAccessor = new iOSTokenCacheAccessor();
            try
            {
                SecStatusCode secStatusCode = iOSTokenCacheAccessor.SaveBrokerApplicationToken(clientId, applicationToken);

                _logger.Info(string.Format(
                   CultureInfo.CurrentCulture,
                   BrokerConstants.AttemptToSaveBrokerApplicationToken + "SecStatusCode: {0}",
                   secStatusCode));
            }
            catch(Exception ex)
            {
                throw new AdalException(
                    AdalErrorIOSEx.WritingApplicationTokenToKeychainFailed,
                    AdalErrorMessageIOSEx.WritingApplicationTokenToKeychainFailed + ex.Message);
            }
        }

        private string TryReadBrokerApplicationTokenFromKeychain(IDictionary<string, string> brokerPaylaod)
        {
            iOSTokenCacheAccessor iOSTokenCacheAccessor = new iOSTokenCacheAccessor();

            try
            {
                SecStatusCode secStatusCode = iOSTokenCacheAccessor.TryGetBrokerApplicationToken(brokerPaylaod[BrokerParameter.ClientId], out string appToken);

                _logger.Info(string.Format(
                   CultureInfo.CurrentCulture,
                   BrokerConstants.SecStatusCodeFromTryGetBrokerApplicationToken + "SecStatusCode: {0}",
                   secStatusCode));

                return appToken;
            }
            catch(Exception ex)
            {
                throw new AdalException(
                   AdalErrorIOSEx.ReadingApplicationTokenFromKeychainFailed,
                   AdalErrorMessageIOSEx.ReadingApplicationTokenFromKeychainFailed + ex.Message);
            }
        }

        public static void SetBrokerResponse(NSUrl responseUrl)
        {
            brokerResponse = responseUrl;
            brokerResponseReady?.Release();
        }
    }
}
