﻿//----------------------------------------------------------------------
//
// Copyright (c) Microsoft Corporation.
// All rights reserved.
//
// This code is licensed under the MIT License.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files(the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions :
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//
//------------------------------------------------------------------------------

using System;
using System.Collections.Generic;
using System.Linq;
using System.Net.Http;
using System.Threading.Tasks;
using Microsoft.Identity.Core;
using Microsoft.Identity.Core.Cache;
using Microsoft.Identity.Core.UI;
using Microsoft.IdentityModel.Clients.ActiveDirectory;
using Microsoft.IdentityModel.Clients.ActiveDirectory.Internal;
using Microsoft.IdentityModel.Clients.ActiveDirectory.Internal.Http;
using Microsoft.IdentityModel.Clients.ActiveDirectory.Internal.OAuth2;
using Microsoft.IdentityModel.Clients.ActiveDirectory.Internal.Platform;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Test.ADAL.NET.Common;
using Test.ADAL.NET.Common.Mocks;

namespace Test.ADAL.NET.Unit
{
#if !NET_CORE // Enable when bug https://IdentityDivision.visualstudio.com/_workitems/edit/573878 is fixed

    /// <summary>
    /// This test class executes and validates OBO scenarios where token cache may or may not 
    /// contain entries with user assertion hash. It accounts for cases where there is
    /// a single user and when there are multiple users in the cache.
    /// user assertion hash exists so that the API can deterministically identify the user
    /// in the cache when a usernae is not passed in. It also allows the API to acquire
    /// new token when a different assertion is passed for the user. this is needed because
    /// the user may have authenticated with updated claims like MFA/device auth on the client.
    /// </summary>
    [TestClass]
    public class OboFlowTests
    {
        private readonly DateTimeOffset _expirationTime = DateTimeOffset.UtcNow + TimeSpan.FromMinutes(30);
        private static readonly string[] _cacheNoise = { "", "different" };
        private ICryptographyManager _crypto;

        [TestInitialize]
        public void TestInitialize()
        {
            ModuleInitializer.ForceModuleInitializationTestOnly();
            InstanceDiscovery.InstanceCache.Clear();

            _crypto = PlatformProxyFactory.GetPlatformProxy().CryptographyManager;
        }

        internal void SetupMocks(MockHttpManager httpManager)
        {
            InstanceDiscovery.InstanceCache.Clear();
            httpManager.AddMockHandler(
                    MockHelpers.CreateInstanceDiscoveryMockHandler(
                        AdalTestConstants.GetDiscoveryEndpoint(
                            AdalTestConstants.DefaultAuthorityCommonTenant)));
        }

        [TestMethod]
        public async Task MultiUserNoHashInCacheNoUsernamePassedInAssertionTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle,
                    AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";

                foreach (var cachenoise in _cacheNoise)
                {
                    //cache entry has no user assertion hash
                    await context.TokenCache.StoreToCacheAsync(new AdalResultWrapper
                    {
                        RefreshToken = cachenoise + "some-rt",
                        ResourceInResponse = AdalTestConstants.DefaultResource,
                        Result = new AdalResult("Bearer", cachenoise + "some-token-in-cache", _expirationTime)
                        {
                            UserInfo =
                                new AdalUserInfo()
                                {
                                    DisplayableId = cachenoise + AdalTestConstants.DefaultDisplayableId,
                                    UniqueId = cachenoise + AdalTestConstants.DefaultUniqueId
                                },
                            IdToken = MockHelpers.CreateAdalIdToken(AdalTestConstants.DefaultUniqueId, AdalTestConstants.DefaultDisplayableId)
                        },
                    },
                    AdalTestConstants.DefaultAuthorityHomeTenant, AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.User,
                   new RequestContext(null, new AdalLogger(new Guid()))).ConfigureAwait(false);
                }

                SetupMocks(httpManager);

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);
                httpManager.AddMockHandler(new MockHttpMessageHandler(AdalTestConstants.GetTokenEndpoint(AdalTestConstants.DefaultAuthorityHomeTenant))
                {
                    Method = HttpMethod.Post,
                    ResponseMessage = MockHelpers.CreateSuccessTokenResponseMessage(),
                    PostData = new Dictionary<string, string>()
                {
                    {"client_id", AdalTestConstants.DefaultClientId},
                    {"grant_type", "urn:ietf:params:oauth:grant-type:jwt-bearer"}
                }
                });

                // call acquire token with no username. this will result in a network call because cache entry with no assertion hash is
                // treated as a cache miss.
                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.DefaultResource, clientCredential,
                            new UserAssertion(accessToken)).ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual("some-access-token", result.AccessToken);
                Assert.AreEqual(AdalTestConstants.DefaultDisplayableId, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(2, context.TokenCache.Count);

                //assertion hash should be stored in the cache entry.
                Assert.AreEqual(_crypto.CreateSha256Hash(accessToken),
                    context.TokenCache._tokenCacheDictionary.Values.First(x => x.UserAssertionHash != null).UserAssertionHash);
            }
        }

        [TestMethod]
        public async Task MultiUserNoHashInCacheMatchingUsernamePassedInAssertionTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle,
                    AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";

                foreach (var cachenoise in _cacheNoise)
                {
                    //cache entry has no user assertion hash
                    await context.TokenCache.StoreToCacheAsync(new AdalResultWrapper
                    {
                        RefreshToken = cachenoise + "some-rt",
                        ResourceInResponse = AdalTestConstants.DefaultResource,
                        Result = new AdalResult("Bearer", cachenoise + "some-token-in-cache", _expirationTime)
                        {
                            UserInfo =
                                new AdalUserInfo()
                                {
                                    DisplayableId = cachenoise + AdalTestConstants.DefaultDisplayableId,
                                    UniqueId = cachenoise + AdalTestConstants.DefaultUniqueId
                                },
                            IdToken = MockHelpers.CreateAdalIdToken(AdalTestConstants.DefaultUniqueId, AdalTestConstants.DefaultDisplayableId)
                        },
                    },
                    AdalTestConstants.DefaultAuthorityHomeTenant, AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.User,
                   new RequestContext(null, new AdalLogger(new Guid()))).ConfigureAwait(false);
                }

                SetupMocks(httpManager);

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);
                httpManager.AddMockHandler(new MockHttpMessageHandler(AdalTestConstants.GetTokenEndpoint(AdalTestConstants.DefaultAuthorityHomeTenant))
                {
                    Method = HttpMethod.Post,
                    ResponseMessage = MockHelpers.CreateSuccessTokenResponseMessage(),
                    PostData = new Dictionary<string, string>()
                {
                    {"client_id", AdalTestConstants.DefaultClientId},
                    {"grant_type", "urn:ietf:params:oauth:grant-type:jwt-bearer"}
                }
                });

                // call acquire token with matching username from cache entry. this will result in a network call 
                // because cache entry with no assertion hash is treated as a cache miss.

                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.DefaultResource, clientCredential,
                            new UserAssertion(accessToken, OAuthGrantType.JwtBearer, AdalTestConstants.DefaultDisplayableId)).ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual("some-access-token", result.AccessToken);
                Assert.AreEqual(AdalTestConstants.DefaultDisplayableId, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(2, context.TokenCache.Count);

                //assertion hash should be stored in the cache entry.
                Assert.AreEqual(_crypto.CreateSha256Hash(accessToken),
                    context.TokenCache._tokenCacheDictionary.Values.First(x => x.UserAssertionHash != null).UserAssertionHash);
            }
        }

        [TestMethod]
        public async Task MultiUserNoHashInCacheDifferentUsernamePassedInAssertionTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle,
                    AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";

                foreach (var cachenoise in _cacheNoise)
                {
                    AdalTokenCacheKey key = new AdalTokenCacheKey(AdalTestConstants.DefaultAuthorityHomeTenant,
                        AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.User,
                        cachenoise + AdalTestConstants.DefaultUniqueId, cachenoise + AdalTestConstants.DefaultDisplayableId);
                    //cache entry has no user assertion hash
                    context.TokenCache._tokenCacheDictionary[key] = new AdalResultWrapper
                    {
                        RefreshToken = cachenoise + "some-rt",
                        ResourceInResponse = AdalTestConstants.DefaultResource,
                        Result = new AdalResult("Bearer", cachenoise + "some-token-in-cache", _expirationTime)
                        {
                            UserInfo =
                                new AdalUserInfo()
                                {
                                    DisplayableId = cachenoise + AdalTestConstants.DefaultDisplayableId,
                                    UniqueId = cachenoise + AdalTestConstants.DefaultUniqueId
                                }
                        },
                    };
                }

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);

                string displayableId2 = "extra" + AdalTestConstants.DefaultDisplayableId;
                string uniqueId2 = "extra" + AdalTestConstants.DefaultUniqueId;

                httpManager.AddMockHandler(new MockHttpMessageHandler(AdalTestConstants.GetTokenEndpoint(AdalTestConstants.DefaultAuthorityHomeTenant))
                {
                    Method = HttpMethod.Post,
                    ResponseMessage =
                        MockHelpers.CreateSuccessTokenResponseMessage(uniqueId2, displayableId2,
                            AdalTestConstants.DefaultResource),
                    PostData = new Dictionary<string, string>()
                {
                    {"client_id", AdalTestConstants.DefaultClientId},
                    {"grant_type", "urn:ietf:params:oauth:grant-type:jwt-bearer"}
                }
                });

                // call acquire token with diferent username from cache entry. this will result in a network call
                // because cache lookup failed for non-existant user
                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.DefaultResource, clientCredential,
                            new UserAssertion(accessToken, OAuthGrantType.JwtBearer,
                                "non-existant" + AdalTestConstants.DefaultDisplayableId)).ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual("some-access-token", result.AccessToken);
                Assert.AreEqual(displayableId2, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(3, context.TokenCache.Count);

                //assertion hash should be stored in the cache entry.
                Assert.AreEqual(_crypto.CreateSha256Hash(accessToken),
                    context.TokenCache._tokenCacheDictionary.Values.First(x => x.UserAssertionHash != null)
                        .UserAssertionHash);
            }
        }

        [TestMethod]
        public async Task MultiUserWithHashInCacheNoUsernameAndMatchingAssertionTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle,
                    AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";
                //cache entry has user assertion hash
                string tokenInCache = "obo-access-token";

                foreach (var cachenoise in _cacheNoise)
                {
                    AdalTokenCacheKey key = new AdalTokenCacheKey(AdalTestConstants.DefaultAuthorityHomeTenant,
                        AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.User,
                        cachenoise + AdalTestConstants.DefaultUniqueId, cachenoise + AdalTestConstants.DefaultDisplayableId);

                    context.TokenCache._tokenCacheDictionary[key] = new AdalResultWrapper
                    {
                        RefreshToken = cachenoise + "some-rt",
                        ResourceInResponse = AdalTestConstants.DefaultResource,
                        Result =
                            new AdalResult("Bearer", cachenoise + tokenInCache, _expirationTime)
                            {
                                UserInfo =
                                    new AdalUserInfo()
                                    {
                                        DisplayableId = cachenoise + AdalTestConstants.DefaultDisplayableId,
                                        UniqueId = cachenoise + AdalTestConstants.DefaultUniqueId
                                    }
                            },
                        UserAssertionHash = _crypto.CreateSha256Hash(cachenoise + accessToken)
                    };
                }

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);

                // call acquire token with no username and matching assertion hash. this will result in a cache
                // hit.
                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.DefaultResource, clientCredential,
                            new UserAssertion(accessToken)).ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual(tokenInCache, result.AccessToken);
                Assert.AreEqual(AdalTestConstants.DefaultDisplayableId, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(2, context.TokenCache.Count);
            }
        }

        [TestMethod]
        public async Task MultiUserWithHashInCacheNoUsernameAndDifferentAssertionTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle,
                    AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";
                //cache entry has user assertion hash
                string tokenInCache = "obo-access-token";

                foreach (var cachenoise in _cacheNoise)
                {
                    await context.TokenCache.StoreToCacheAsync(new AdalResultWrapper
                    {
                        RefreshToken = cachenoise + "some-rt",
                        ResourceInResponse = AdalTestConstants.DefaultResource,
                        Result =
                            new AdalResult("Bearer", cachenoise + tokenInCache, _expirationTime)
                            {
                                UserInfo =
                                    new AdalUserInfo()
                                    {
                                        DisplayableId = cachenoise + AdalTestConstants.DefaultDisplayableId,
                                        UniqueId = cachenoise + AdalTestConstants.DefaultUniqueId
                                    },
                                IdToken = MockHelpers.CreateAdalIdToken(AdalTestConstants.DefaultUniqueId, AdalTestConstants.DefaultDisplayableId)
                            },
                        UserAssertionHash = _crypto.CreateSha256Hash(cachenoise + accessToken)
                    },
                    AdalTestConstants.DefaultAuthorityHomeTenant, AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.User,
                   new RequestContext(null, new AdalLogger(new Guid()))).ConfigureAwait(false);
                }

                SetupMocks(httpManager);

                httpManager.AddMockHandler(new MockHttpMessageHandler(AdalTestConstants.GetTokenEndpoint(AdalTestConstants.DefaultAuthorityHomeTenant))
                {
                    Method = HttpMethod.Post,
                    ResponseMessage = MockHelpers.CreateSuccessTokenResponseMessage(),
                    PostData = new Dictionary<string, string>()
                {
                    {"client_id", AdalTestConstants.DefaultClientId},
                    {"grant_type", "urn:ietf:params:oauth:grant-type:jwt-bearer"}
                }
                });

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);

                // call acquire token with no username and different assertion hash. this will result in a cache
                // hit.
                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.DefaultResource, clientCredential,
                            new UserAssertion("non-existant" + accessToken)).ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual("some-access-token", result.AccessToken);
                Assert.AreEqual(AdalTestConstants.DefaultDisplayableId, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(2, context.TokenCache.Count);
            }
        }

        [TestMethod]
        public async Task MultiUserWithHashInCacheMatchingUsernameAndMatchingAssertionTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle,
                    AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";
                //cache entry has user assertion hash
                string tokenInCache = "obo-access-token";
                foreach (var cachenoise in _cacheNoise)
                {
                    AdalTokenCacheKey key = new AdalTokenCacheKey(AdalTestConstants.DefaultAuthorityHomeTenant,
                        AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.User,
                        cachenoise + AdalTestConstants.DefaultUniqueId, cachenoise + AdalTestConstants.DefaultDisplayableId);

                    context.TokenCache._tokenCacheDictionary[key] = new AdalResultWrapper
                    {
                        RefreshToken = cachenoise + "some-rt",
                        ResourceInResponse = AdalTestConstants.DefaultResource,
                        Result =
                            new AdalResult("Bearer", cachenoise + tokenInCache, _expirationTime)
                            {
                                UserInfo =
                                    new AdalUserInfo()
                                    {
                                        DisplayableId = cachenoise + AdalTestConstants.DefaultDisplayableId,
                                        UniqueId = cachenoise + AdalTestConstants.DefaultUniqueId
                                    }
                            },
                        UserAssertionHash = _crypto.CreateSha256Hash(cachenoise + accessToken)
                    };
                }

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);

                // call acquire token with matching username and matching assertion hash. this will result in a cache
                // hit.
                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.DefaultResource, clientCredential,
                            new UserAssertion(accessToken, OAuthGrantType.JwtBearer, AdalTestConstants.DefaultDisplayableId)).ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual(tokenInCache, result.AccessToken);
                Assert.AreEqual(AdalTestConstants.DefaultDisplayableId, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(2, context.TokenCache.Count);

                //assertion hash should be stored in the cache entry.
                var expectedHash = _crypto.CreateSha256Hash(accessToken);

                Assert.IsTrue(context.TokenCache._tokenCacheDictionary.Values.Any(v => v.UserAssertionHash == expectedHash));
            }
        }

        [TestMethod]
        public async Task MultiUserWithHashInCacheMatchingUsernameAndDifferentAssertionTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle,
                    AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";
                //cache entry has user assertion hash
                string tokenInCache = "obo-access-token";
                foreach (var cachenoise in _cacheNoise)
                {
                    await context.TokenCache.StoreToCacheAsync(new AdalResultWrapper
                    {
                        RefreshToken = cachenoise + "some-rt",
                        ResourceInResponse = AdalTestConstants.DefaultResource,
                        Result =
                            new AdalResult("Bearer", cachenoise + tokenInCache, _expirationTime)
                            {
                                UserInfo =
                                    new AdalUserInfo()
                                    {
                                        DisplayableId = cachenoise + AdalTestConstants.DefaultDisplayableId,
                                        UniqueId = cachenoise + AdalTestConstants.DefaultUniqueId
                                    },
                                IdToken = MockHelpers.CreateAdalIdToken(AdalTestConstants.DefaultUniqueId, AdalTestConstants.DefaultDisplayableId)
                            },
                        UserAssertionHash = _crypto.CreateSha256Hash(cachenoise + accessToken)
                    },
                    AdalTestConstants.DefaultAuthorityHomeTenant, AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.User,
                   new RequestContext(null, new AdalLogger(new Guid()))).ConfigureAwait(false);
                }

                SetupMocks(httpManager);
                httpManager.AddMockHandler(new MockHttpMessageHandler(AdalTestConstants.GetTokenEndpoint(AdalTestConstants.DefaultAuthorityHomeTenant))
                {
                    Method = HttpMethod.Post,
                    ResponseMessage = MockHelpers.CreateSuccessTokenResponseMessage(),
                    PostData = new Dictionary<string, string>()
                {
                    {"client_id", AdalTestConstants.DefaultClientId},
                    {"grant_type", "urn:ietf:params:oauth:grant-type:jwt-bearer"}
                }
                });

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);

                // call acquire token with matching username and different assertion hash. this will result in a cache
                // hit.
                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.DefaultResource, clientCredential,
                            new UserAssertion("non-existant" + accessToken, OAuthGrantType.JwtBearer, AdalTestConstants.DefaultDisplayableId))
                        .ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual("some-access-token", result.AccessToken);
                Assert.AreEqual(AdalTestConstants.DefaultDisplayableId, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(2, context.TokenCache.Count);
            }
        }

        [TestMethod]
        public async Task SingleUserNoHashInCacheNoUsernamePassedInAssertionTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle,
                    AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";
                await context.TokenCache.StoreToCacheAsync(new AdalResultWrapper
                {
                    RefreshToken = "some-rt",
                    ResourceInResponse = AdalTestConstants.DefaultResource,
                    Result = new AdalResult("Bearer", "some-token-in-cache", _expirationTime)
                    {
                        UserInfo =
                            new AdalUserInfo()
                            {
                                DisplayableId = AdalTestConstants.DefaultDisplayableId,
                                UniqueId = AdalTestConstants.DefaultUniqueId
                            },
                        IdToken = MockHelpers.CreateAdalIdToken(AdalTestConstants.DefaultUniqueId, AdalTestConstants.DefaultDisplayableId)
                    },
                    //cache entry has no user assertion hash
                },
                AdalTestConstants.DefaultAuthorityHomeTenant, AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.User,
               new RequestContext(null, new AdalLogger(new Guid()))).ConfigureAwait(false);

                SetupMocks(httpManager);

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);
                httpManager.AddMockHandler(new MockHttpMessageHandler(AdalTestConstants.GetTokenEndpoint(AdalTestConstants.DefaultAuthorityHomeTenant))
                {
                    Method = HttpMethod.Post,
                    ResponseMessage = MockHelpers.CreateSuccessTokenResponseMessage(),
                    PostData = new Dictionary<string, string>()
                {
                    {"client_id", AdalTestConstants.DefaultClientId},
                    {"grant_type", "urn:ietf:params:oauth:grant-type:jwt-bearer"}
                }
                });

                // call acquire token with no username. this will result in a network call because cache entry with no assertion hash is
                // treated as a cache miss.
                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.DefaultResource, clientCredential,
                            new UserAssertion(accessToken)).ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual("some-access-token", result.AccessToken);
                Assert.AreEqual(AdalTestConstants.DefaultDisplayableId, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(1, context.TokenCache.Count);

                //assertion hash should be stored in the cache entry.
                Assert.AreEqual(_crypto.CreateSha256Hash(accessToken),
                    context.TokenCache._tokenCacheDictionary.Values.First().UserAssertionHash);
            }
        }

        [TestMethod]
        public async Task SingleUserNoHashInCacheMatchingUsernamePassedInAssertionTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle,
                    AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";
                await context.TokenCache.StoreToCacheAsync(new AdalResultWrapper
                {
                    RefreshToken = "some-rt",
                    ResourceInResponse = AdalTestConstants.DefaultResource,
                    Result = new AdalResult("Bearer", "some-token-in-cache", _expirationTime)
                    {
                        UserInfo =
                            new AdalUserInfo()
                            {
                                DisplayableId = AdalTestConstants.DefaultDisplayableId,
                                UniqueId = AdalTestConstants.DefaultUniqueId
                            },
                        IdToken = MockHelpers.CreateAdalIdToken(AdalTestConstants.DefaultUniqueId, AdalTestConstants.DefaultDisplayableId)
                    },
                    //cache entry has no user assertion hash
                },
                AdalTestConstants.DefaultAuthorityHomeTenant, AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.User,
               new RequestContext(null, new AdalLogger(new Guid()))).ConfigureAwait(false);

                SetupMocks(httpManager);

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);
                httpManager.AddMockHandler(new MockHttpMessageHandler(AdalTestConstants.GetTokenEndpoint(AdalTestConstants.DefaultAuthorityHomeTenant))
                {
                    Method = HttpMethod.Post,
                    ResponseMessage = MockHelpers.CreateSuccessTokenResponseMessage(),
                    PostData = new Dictionary<string, string>()
                {
                    {"client_id", AdalTestConstants.DefaultClientId},
                    {"grant_type", "urn:ietf:params:oauth:grant-type:jwt-bearer"}
                }
                });

                // call acquire token with matching username from cache entry. this will result in a network call 
                // because cache entry with no assertion hash is treated as a cache miss.

                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.DefaultResource, clientCredential,
                            new UserAssertion(accessToken, OAuthGrantType.JwtBearer, AdalTestConstants.DefaultDisplayableId)).ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual("some-access-token", result.AccessToken);
                Assert.AreEqual(AdalTestConstants.DefaultDisplayableId, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(1, context.TokenCache.Count);

                //assertion hash should be stored in the cache entry.
                Assert.AreEqual(_crypto.CreateSha256Hash(accessToken),
                    context.TokenCache._tokenCacheDictionary.Values.First().UserAssertionHash);
            }
        }

        [TestMethod]
        public async Task SingleUserNoHashInCacheDifferentUsernamePassedInAssertionTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle,
                    AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";
                AdalTokenCacheKey key = new AdalTokenCacheKey(AdalTestConstants.DefaultAuthorityHomeTenant,
                    AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.User,
                    AdalTestConstants.DefaultUniqueId, AdalTestConstants.DefaultDisplayableId);
                //cache entry has no user assertion hash
                context.TokenCache._tokenCacheDictionary[key] = new AdalResultWrapper
                {
                    RefreshToken = "some-rt",
                    ResourceInResponse = AdalTestConstants.DefaultResource,
                    Result = new AdalResult("Bearer", "some-token-in-cache", _expirationTime)
                    {
                        UserInfo =
                            new AdalUserInfo()
                            {
                                DisplayableId = AdalTestConstants.DefaultDisplayableId,
                                UniqueId = AdalTestConstants.DefaultUniqueId
                            }
                    },
                };

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);

                string displayableId2 = "extra" + AdalTestConstants.DefaultDisplayableId;
                string uniqueId2 = "extra" + AdalTestConstants.DefaultUniqueId;

                httpManager.AddMockHandler(new MockHttpMessageHandler(AdalTestConstants.GetTokenEndpoint(AdalTestConstants.DefaultAuthorityHomeTenant))
                {
                    Method = HttpMethod.Post,
                    ResponseMessage =
                        MockHelpers.CreateSuccessTokenResponseMessage(uniqueId2, displayableId2,
                            AdalTestConstants.DefaultResource),
                    PostData = new Dictionary<string, string>()
                {
                    {"client_id", AdalTestConstants.DefaultClientId},
                    {"grant_type", "urn:ietf:params:oauth:grant-type:jwt-bearer"}
                }
                });

                // call acquire token with diferent username from cache entry. this will result in a network call
                // because cache lookup failed for non-existant user
                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.DefaultResource, clientCredential,
                            new UserAssertion(accessToken, OAuthGrantType.JwtBearer, displayableId2
                                )).ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual("some-access-token", result.AccessToken);
                Assert.AreEqual(displayableId2, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(2, context.TokenCache.Count);

                //assertion hash should be stored in the cache entry.
                Assert.AreEqual(_crypto.CreateSha256Hash(accessToken),
                    context.TokenCache._tokenCacheDictionary.Values.First(
                        s => s.Result.UserInfo != null && s.Result.UserInfo.DisplayableId.Equals(displayableId2, StringComparison.OrdinalIgnoreCase))
                        .UserAssertionHash);
            }
        }

        [TestMethod]
        public async Task SingleUserWithHashInCacheNoUsernameAndMatchingAssertionTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle, AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";
                AdalTokenCacheKey key = new AdalTokenCacheKey(AdalTestConstants.DefaultAuthorityHomeTenant,
                    AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.User,
                    AdalTestConstants.DefaultUniqueId, AdalTestConstants.DefaultDisplayableId);

                //cache entry has user assertion hash
                string tokenInCache = "obo-access-token";
                context.TokenCache._tokenCacheDictionary[key] = new AdalResultWrapper
                {
                    RefreshToken = "some-rt",
                    ResourceInResponse = AdalTestConstants.DefaultResource,
                    Result =
                        new AdalResult("Bearer", tokenInCache, _expirationTime)
                        {
                            UserInfo =
                                new AdalUserInfo()
                                {
                                    DisplayableId = AdalTestConstants.DefaultDisplayableId,
                                    UniqueId = AdalTestConstants.DefaultUniqueId
                                }
                        },
                    UserAssertionHash = _crypto.CreateSha256Hash(accessToken)
                };

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);

                // call acquire token with no username and matching assertion hash. this will result in a cache
                // hit.
                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.DefaultResource, clientCredential,
                            new UserAssertion(accessToken)).ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual(tokenInCache, result.AccessToken);
                Assert.AreEqual(AdalTestConstants.DefaultDisplayableId, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(1, context.TokenCache.Count);

                //assertion hash should be stored in the cache entry.
                Assert.AreEqual(_crypto.CreateSha256Hash(accessToken),
                    context.TokenCache._tokenCacheDictionary.Values.First().UserAssertionHash);
            }
        }

        [TestMethod]
        public async Task SingleUserWithHashInCacheNoUsernameAndDifferentAssertionTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle,
                    AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";
                string tokenInCache = "obo-access-token";
                await context.TokenCache.StoreToCacheAsync(new AdalResultWrapper
                {
                    RefreshToken = "some-rt",
                    ResourceInResponse = AdalTestConstants.DefaultResource,
                    Result =
                        new AdalResult("Bearer", tokenInCache, _expirationTime)
                        {
                            UserInfo =
                                new AdalUserInfo()
                                {
                                    DisplayableId = AdalTestConstants.DefaultDisplayableId,
                                    UniqueId = AdalTestConstants.DefaultUniqueId
                                },
                            IdToken = MockHelpers.CreateAdalIdToken(AdalTestConstants.DefaultUniqueId, AdalTestConstants.DefaultDisplayableId)
                        },
                    UserAssertionHash = _crypto.CreateSha256Hash(accessToken + "different")
                },
                AdalTestConstants.DefaultAuthorityHomeTenant, AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.User,
               new RequestContext(null, new AdalLogger(new Guid()))).ConfigureAwait(false);

                SetupMocks(httpManager);

                httpManager.AddMockHandler(new MockHttpMessageHandler(AdalTestConstants.GetTokenEndpoint(AdalTestConstants.DefaultAuthorityHomeTenant))
                {
                    Method = HttpMethod.Post,
                    ResponseMessage = MockHelpers.CreateSuccessTokenResponseMessage(),
                    PostData = new Dictionary<string, string>()
                {
                    {"client_id", AdalTestConstants.DefaultClientId},
                    {"grant_type", "urn:ietf:params:oauth:grant-type:jwt-bearer"}
                }
                });

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);

                // call acquire token with no username and different assertion hash. this will result in a cache
                // hit.
                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.DefaultResource, clientCredential,
                            new UserAssertion(accessToken)).ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual("some-access-token", result.AccessToken);
                Assert.AreEqual(AdalTestConstants.DefaultDisplayableId, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(1, context.TokenCache.Count);

                //assertion hash should be stored in the cache entry.
                Assert.AreEqual(_crypto.CreateSha256Hash(accessToken),
                    context.TokenCache._tokenCacheDictionary.Values.First().UserAssertionHash);
            }
        }

        [TestMethod]
        public async Task SingleUserWithHashInCacheMatchingUsernameAndMatchingAssertionTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle,
                    AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";
                AdalTokenCacheKey key = new AdalTokenCacheKey(AdalTestConstants.DefaultAuthorityHomeTenant,
                    AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.User,
                    AdalTestConstants.DefaultUniqueId, AdalTestConstants.DefaultDisplayableId);

                //cache entry has user assertion hash
                string tokenInCache = "obo-access-token";
                context.TokenCache._tokenCacheDictionary[key] = new AdalResultWrapper
                {
                    RefreshToken = "some-rt",
                    ResourceInResponse = AdalTestConstants.DefaultResource,
                    Result =
                        new AdalResult("Bearer", tokenInCache, _expirationTime)
                        {
                            UserInfo =
                                new AdalUserInfo()
                                {
                                    DisplayableId = AdalTestConstants.DefaultDisplayableId,
                                    UniqueId = AdalTestConstants.DefaultUniqueId
                                }
                        },
                    UserAssertionHash = _crypto.CreateSha256Hash(accessToken)
                };

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);

                // call acquire token with matching username and matching assertion hash. this will result in a cache
                // hit.
                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.DefaultResource, clientCredential,
                            new UserAssertion(accessToken)).ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual(tokenInCache, result.AccessToken);
                Assert.AreEqual(AdalTestConstants.DefaultDisplayableId, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(1, context.TokenCache.Count);

                //assertion hash should be stored in the cache entry.
                Assert.AreEqual(_crypto.CreateSha256Hash(accessToken),
                    context.TokenCache._tokenCacheDictionary.Values.First().UserAssertionHash);
            }
        }

        [TestMethod]
        public async Task SingleUserWithHashInCacheMatchingUsernameAndDifferentAssertionTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle,
                    AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";
                string tokenInCache = "obo-access-token";
                await context.TokenCache.StoreToCacheAsync(new AdalResultWrapper
                {
                    RefreshToken = "some-rt",
                    ResourceInResponse = AdalTestConstants.DefaultResource,
                    Result =
                        new AdalResult("Bearer", tokenInCache, _expirationTime)
                        {
                            UserInfo =
                                new AdalUserInfo()
                                {
                                    DisplayableId = AdalTestConstants.DefaultDisplayableId,
                                    UniqueId = AdalTestConstants.DefaultUniqueId
                                },
                            IdToken = MockHelpers.CreateAdalIdToken(AdalTestConstants.DefaultUniqueId, AdalTestConstants.DefaultDisplayableId)
                        },
                    UserAssertionHash = _crypto.CreateSha256Hash(accessToken + "different")
                },
                AdalTestConstants.DefaultAuthorityHomeTenant, AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.User,
               new RequestContext(null, new AdalLogger(new Guid()))).ConfigureAwait(false);

                SetupMocks(httpManager);

                httpManager.AddMockHandler(new MockHttpMessageHandler(AdalTestConstants.GetTokenEndpoint(AdalTestConstants.DefaultAuthorityHomeTenant))
                {
                    Method = HttpMethod.Post,
                    ResponseMessage = MockHelpers.CreateSuccessTokenResponseMessage(),
                    PostData = new Dictionary<string, string>()
                {
                    {"client_id", AdalTestConstants.DefaultClientId},
                    {"grant_type", "urn:ietf:params:oauth:grant-type:jwt-bearer"}
                }
                });

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);

                // call acquire token with matching username and different assertion hash. this will result in a cache
                // hit.
                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.DefaultResource, clientCredential,
                            new UserAssertion(accessToken)).ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual("some-access-token", result.AccessToken);
                Assert.AreEqual(AdalTestConstants.DefaultDisplayableId, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(1, context.TokenCache.Count);

                //assertion hash should be stored in the cache entry.
                Assert.AreEqual(_crypto.CreateSha256Hash(accessToken),
                    context.TokenCache._tokenCacheDictionary.Values.First().UserAssertionHash);
            }
        }

        [TestMethod]
        public async Task SingleUserWithHashInCacheMatchingUsernameAndMatchingAssertionDifferentResourceTestAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                var context = new AuthenticationContext(
                    serviceBundle,
                    AdalTestConstants.DefaultAuthorityHomeTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                string accessToken = "access-token";
                AdalTokenCacheKey key = new AdalTokenCacheKey(AdalTestConstants.DefaultAuthorityHomeTenant,
                    AdalTestConstants.DefaultResource, AdalTestConstants.DefaultClientId, TokenSubjectType.UserPlusClient,
                    AdalTestConstants.DefaultUniqueId, AdalTestConstants.DefaultDisplayableId);

                //cache entry has user assertion hash
                string tokenInCache = "obo-access-token";
                context.TokenCache._tokenCacheDictionary[key] = new AdalResultWrapper
                {
                    RefreshToken = "some-rt",
                    ResourceInResponse = AdalTestConstants.DefaultResource,
                    Result =
                        new AdalResult("Bearer", tokenInCache, _expirationTime)
                        {
                            UserInfo =
                                new AdalUserInfo()
                                {
                                    DisplayableId = AdalTestConstants.DefaultDisplayableId,
                                    UniqueId = AdalTestConstants.DefaultUniqueId
                                }
                        },
                    UserAssertionHash = _crypto.CreateSha256Hash(accessToken)
                };

                httpManager.AddMockHandler(new MockHttpMessageHandler(AdalTestConstants.GetTokenEndpoint(AdalTestConstants.DefaultAuthorityHomeTenant))
                {
                    Method = HttpMethod.Post,
                    ResponseMessage = MockHelpers.CreateSuccessTokenResponseMessage(AdalTestConstants.AnotherResource,
                    AdalTestConstants.DefaultDisplayableId, AdalTestConstants.DefaultUniqueId),
                    PostData = new Dictionary<string, string>()
                {
                    {"client_id", AdalTestConstants.DefaultClientId},
                    {"grant_type", "urn:ietf:params:oauth:grant-type:jwt-bearer"}
                }
                });

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                    AdalTestConstants.DefaultClientSecret);

                // call acquire token with matching username and different assertion hash. this will result in a cache
                // hit.
                var result =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.AnotherResource, clientCredential,
                            new UserAssertion(accessToken, OAuthGrantType.JwtBearer, AdalTestConstants.DefaultDisplayableId)).ConfigureAwait(false);
                Assert.IsNotNull(result.AccessToken);
                Assert.AreEqual("some-access-token", result.AccessToken);
                Assert.AreEqual(AdalTestConstants.DefaultDisplayableId, result.UserInfo.DisplayableId);

                //there should be only one cache entry.
                Assert.AreEqual(2, context.TokenCache.Count);

                //assertion hash should be stored in the cache entry.
                foreach (var value in context.TokenCache._tokenCacheDictionary.Values)
                {
                    Assert.AreEqual(_crypto.CreateSha256Hash(accessToken), value.UserAssertionHash);
                }
            }
        }

        [TestMethod]
        public async Task OboATForAT2CacheLookUp_OboATForAT1Cached_CacheMissAsync()
        {
            using (var httpManager = new MockHttpManager())
            {
                AuthenticationContext context = null;
                var serviceBundle = ServiceBundle.CreateWithCustomHttpManager(httpManager);

                SetupMocks(httpManager);

                httpManager.AddMockHandler(new MockHttpMessageHandler(AdalTestConstants.GetTokenEndpoint(AdalTestConstants.DefaultAuthorityCommonTenant))
                {
                    Method = HttpMethod.Post,
                    ResponseMessage = MockHelpers.CreateSuccessTokenResponseMessage(AdalTestConstants.AnotherResource,
                    AdalTestConstants.DefaultDisplayableId, AdalTestConstants.DefaultUniqueId),
                    PostData = new Dictionary<string, string>()
                {
                    {"client_id", AdalTestConstants.DefaultClientId},
                    {"grant_type", "urn:ietf:params:oauth:grant-type:jwt-bearer"}
                }
                });

                context = new AuthenticationContext(
                    serviceBundle, AdalTestConstants.DefaultAuthorityCommonTenant,
                    AuthorityValidationType.NotProvided,
                    new TokenCache());

                ClientCredential clientCredential = new ClientCredential(AdalTestConstants.DefaultClientId,
                        AdalTestConstants.DefaultClientSecret);

                var result1 = await context.AcquireTokenAsync
                    (AdalTestConstants.AnotherResource, clientCredential, new UserAssertion("Access Token1", OAuthGrantType.JwtBearer))
                    .ConfigureAwait(false);

                Assert.IsNotNull(result1);
                Assert.AreEqual(result1.AccessToken, "some-access-token");
                Assert.IsNotNull(result1.UserInfo);

                httpManager.AddMockHandler(new MockHttpMessageHandler(AdalTestConstants.GetTokenEndpoint(AdalTestConstants.TenantSpecificAuthority))
                {
                    Method = HttpMethod.Post,
                    ResponseMessage = MockHelpers.CreateSuccessTokenResponseMessage(AdalTestConstants.AnotherResource,
                    AdalTestConstants.DefaultDisplayableId + "_1", AdalTestConstants.DefaultUniqueId + "_1"),
                    PostData = new Dictionary<string, string>()
                {
                    {"client_id", AdalTestConstants.DefaultClientId},
                    {"grant_type", "urn:ietf:params:oauth:grant-type:jwt-bearer"}
                }
                });

                // make sure obo grant is used, so matching AT or RT is not found in the cache
                var result2 =
                    await
                        context.AcquireTokenAsync(AdalTestConstants.AnotherResource, clientCredential,
                            new UserAssertion("Access Token2", OAuthGrantType.JwtBearer)).ConfigureAwait(false);

                Assert.IsNotNull(result2.AccessToken);
            }
        }
    }
#endif
}